require 'wavefile/chunk_readers/riff_reader'
require 'wavefile/chunk_readers/base_chunk_reader'
require 'wavefile/chunk_readers/format_chunk_reader'
require 'wavefile/chunk_readers/data_chunk_reader'
require 'wavefile/chunk_readers/sample_chunk_reader'

module WaveFile
  # Internal
  module ChunkReaders    # :nodoc:
  end
end
